`timescale 1ns / 100ps

module m68k(clk,
	    fc, a, _as, _uds, _lds, r_w, _dtack, d,
	    _ipl, _avec,
	    _br, _bg,
	    _berr, _reset, _halt);

	/* Clock */
	input			clk;

	/* Bus interface */
	output reg [23:1] 	a;	
	output reg [2:0] 	fc;
	output reg		_as;
	output reg		_uds;
	output reg 		_lds;	
	output reg		r_w;
	input 			_dtack;
 	inout [15:0] 		d;

	/* Bus master control */
	input			_br;
	output reg		_bg;	
 	
	/* Interrupt */
	input [2:0] 		_ipl;
	input 			_avec;

	/* System control */
	input			_berr;
 	input			_reset;
 	input 			_halt;

	integer	 		emu_state;
	integer			bus_state;

	localparam log_none	= 0;
	localparam log_prim_op	= 'h1;
	localparam log_disasm	= 'h2; /* ignore log_prim_op when set */
	localparam log_except	= 'h4;
	localparam log_undef	= 'h8;
	localparam log_all	= 'hff;

	parameter log_flags	= log_none;	
	parameter turbo		= 0;	
	
	localparam emu_reset		= 0;
	localparam emu_running		= 1;
	localparam emu_reading		= 2;
	localparam emu_writing		= 3;
	localparam emu_iack		= 4;
	localparam emu_reading_done	= 5;
	localparam emu_writing_done	= 6;

	/* Reminder:
	 *
	 * FC2   FC1   FC0
	 *  0     0     0    reserved
	 *  0     0     1    user data
	 *  0     1     0    user program
	 *  0     1     1    reserved
	 *  1     0     0    reserved
	 *  1     0     1    supervisor data
	 *  1     1     0    supervisor program
	 *  1     1     1    cpu space
	 * 
	 * (yeah I know we only do 110 for now..)
	 */
	integer 		em;
	integer			em_addr;
	integer			em_data;
	integer			em_flags;

	reg [2:0]		cycle_state; /* For debug */

	reg			d_out;
	reg [15:0]		d_buf;  

	assign d = (d_out) ? d_buf : 16'bz;

	initial begin
		em = $m68k_create(1/*turbo*/, log_flags);
		emu_state = emu_reset;
		fc[2:0] = 3'b000;
		a = 'bz;
		_as = 1;
		_uds = 1;
		_lds = 1;
		r_w = 1;
		_bg = 1;
		d_out = 0;
		d_buf = 16'b1;
		cycle_state = -1;		
	end

   	always@(_ipl) begin
	   $m68k_set_ipl(em, ~_ipl);	   
	end
   
	always@(posedge clk or _reset) begin
		/* Out of cycle state, A and D busses in hi-z */
		d_out = 0;
		a = 'bz;
		if (_reset == 1) begin
			emu_state = $m68k_tick(em, em_addr, em_data, em_flags);
			case (emu_state)
				emu_reading: begin
					/* S0 */
					@(posedge clk);
					cycle_state = 0;
					fc[2:0] = 3'b110;
					r_w = 1;
					/* S1 */
					@(negedge clk);
					cycle_state = 1;
					a[23:1] = em_addr[23:1];
					/* S2 */
					@(posedge clk);
					cycle_state = 2;
					_as = 0;
					_uds = !(em_flags & 2);
					_lds = !(em_flags & 1);
					/* S3 */
					@(negedge clk);
					cycle_state = 3;
					/* S4 */
					@(posedge clk);
					cycle_state = 4;
					@(negedge clk);
					while (_dtack && _berr && _reset)
						@(negedge clk);
					/* S5 */
					cycle_state = 5;
					/* S6 */
					@(posedge clk);
					cycle_state = 6;
					/* S7 */
					@(negedge clk);
					cycle_state = 7;
					/* XXX data should be latched at end of cycle
					 * but I'm having problems with iverilog, so
					 * stick to latching now.
					 */
					em_data = d;
					_as = 1;
					_uds = 1;
					_lds = 1;
//					$display("read done addr %x data %x\n",
//						 em_addr, em_data);
					
					$m68k_read_done(em, em_data);
				end
				emu_writing: begin
					/* S0 */
					@(posedge clk);
					cycle_state = 0;
					fc[2:0] = 3'b110;
					d_buf = em_data;
					r_w = 1;
					/* S1 */
					@(negedge clk);
					cycle_state = 1;
					a[23:1] = em_addr[23:1];
					/* S2 */
					@(posedge clk);
					cycle_state = 2;
					_as = 0;
					r_w = 0;
					/* S3 */
					@(negedge clk);
					cycle_state = 3;
					d_out = 1;
					/* S4 */
					@(posedge clk);
					cycle_state = 4;
					_uds = !(em_flags & 2);
					_lds = !(em_flags & 1);
					@(negedge clk);
					while (_dtack && _berr && _reset)
						@(negedge clk);
					/* S5 */
					cycle_state = 5;
					/* S6 */
					@(posedge clk);
					cycle_state = 6;
					/* S7 */
					@(negedge clk);
					cycle_state = 7;
					_as = 1;
					_uds = 1;
					_lds = 1;
					r_w = 1;
					$m68k_write_done(em);
				end
			endcase
		end
	end
endmodule
