#include <iverilog/vpi_user.h>
#include <pthread.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "e68000/e68000.h"

/* Emu state */
enum emu_state {
	emu_reset,	/* Emulator reset */
	emu_running,	/* Emulator running */
	emu_reading,	/* Reading bus data */
	emu_writing,	/* Writing bus data */
	emu_iack,	/* Interrupt ack */
	emu_reading_done,
	emu_writing_done,
};

/* Cond returns */
enum {
	cond_read_done,
	cond_write_done
};

/* Log flags */
enum {
	log_none	= 0,
	log_prim_op	= 0x1,
	log_disasm	= 0x2, /* ignore log_prim_op when set */
	log_except	= 0x4,
	log_undef	= 0x8,
	log_all		= 0xff,
};

struct m68k_instance {
	/* Emulation state */
	e68000_t	*e68;
	pthread_t	thread;
	pthread_mutex_t	mutex;
	pthread_cond_t	cond_run;
	pthread_cond_t	cond_stop;
	enum emu_state	state;
	int		tick;
	int		log_flags;
	uint32_t	addr;	/* For bus accesses */
	uint16_t	data;	/* For bus accesses */
	int		flags;	/* 0x2 = UDS, 0x1 = LDS */
};

static struct m68k_instance *m68k;

static struct m68k_instance *get_instance(int idx)
{
	/* We only support one instance for now */
	if (idx == 0x1234)
		return m68k;
	vpi_printf("m68k: Bad instance %d.", idx);
	vpi_control(vpiFinish, 1);
	return NULL;
}

static unsigned char get_uint8(void *ext, unsigned long addr)
{
	struct m68k_instance *em = ext;

	assert(em->state == emu_running);

	em->state = emu_reading;
	em->addr = addr & ~1ul;
	em->flags = (addr & 1) ? 0x1 : 0x2;
	while(em->state != emu_reading_done) {
		pthread_cond_signal(&em->cond_stop);
		pthread_cond_wait(&em->cond_run, &em->mutex);
	}
	em->state = emu_running;
	return (addr & 1) ? (em->data & 0xff) : ((em->data >> 8) & 0xff);
}

static unsigned short get_uint16(void *ext, unsigned long addr)
{
	struct m68k_instance *em = ext;

	assert(em->state == emu_running);

	/* Assume aligned */
	em->state = emu_reading;
	em->addr = addr & ~1ul;
	em->flags = 0x3;
	while(em->state != emu_reading_done) {
		pthread_cond_signal(&em->cond_stop);
		pthread_cond_wait(&em->cond_run, &em->mutex);
	}
	em->state = emu_running;
	return em->data;
}

static unsigned long get_uint32(void *ext, unsigned long addr)
{
	unsigned long data;

	data = get_uint16(ext, addr);
	data <<= 16;
	data |= get_uint16(ext, addr + 2);

	return data;
}

static void set_uint8(void *ext, unsigned long addr, unsigned char val)
{
	struct m68k_instance *em = ext;

	assert(em->state == emu_running);

	em->state = emu_writing;
	em->addr = addr & ~1ul;
	em->data = (((uint16_t)val) << 8) | val;
	em->flags = (addr & 1) ? 0x1 : 0x2;
	while(em->state != emu_writing_done) {
		pthread_cond_signal(&em->cond_stop);
		pthread_cond_wait(&em->cond_run, &em->mutex);
	}
	em->state = emu_running;
}

static void set_uint16(void *ext, unsigned long addr, unsigned short val)
{
	struct m68k_instance *em = ext;

	assert(em->state == emu_running);

	em->state = emu_writing;
	em->addr = addr & ~1ul;
	em->data = val;
	em->flags = 0x3;
	while(em->state != emu_writing_done) {
		pthread_cond_signal(&em->cond_stop);
		pthread_cond_wait(&em->cond_run, &em->mutex);
	}
	em->state = emu_running;
}

static void set_uint32(void *ext, unsigned long addr, unsigned long val)
{
	set_uint16(ext, addr, val >> 16);
	set_uint16(ext, addr + 2, val & 0xffff);
}

static void m68k_dasm_str(char *dst, e68_dasm_t *op)
{
	unsigned   i, n;
	char       tmp[256];
	const char *ins;

	strcpy (dst, "");

	for (i = 0; i < op->irn; i++) {
		sprintf (tmp, "%04X ", (unsigned) op->ir[i]);
		strcat (dst, tmp);
	}

	for (i = op->irn; i < 4; i++) {
		strcat (dst, "     ");
	}

	if (op->flags & E68_DFLAG_PRIV) {
		strcat (dst, "*");
	}
	else if (op->flags & E68_DFLAG_CALL) {
		strcat (dst, ">");
	}
	else if (op->flags & E68_DFLAG_RTS) {
		strcat (dst, "<");
	}
	else {
		strcat (dst, " ");
	}

#if 0
      	ins = mac_get_trap_name (op->ir[0]);

	if (ins == NULL) {
		ins = op->op;
	}
#else
		ins = op->op;
#endif

	switch (op->argn) {
	case 0:
		sprintf (tmp, "%s", ins);
		break;

	case 1:
		sprintf (tmp, "%-8s %s", ins, op->arg1);
		break;

	case 2:
		sprintf (tmp, "%-8s %s, %s", ins, op->arg1, op->arg2);
		break;

	case 3:
		sprintf (tmp, "%-8s %s, %s, %s",
			ins, op->arg1, op->arg2, op->arg3
		);
		break;

	default:
		strcpy (tmp, "---");
		break;
	}

	strcat (dst, tmp);

	if (op->comm[0] != 0) {
		n = strlen (dst);

		dst += n;

		while (n < 50) {
			*(dst++) = ' ';
			n += 1;
		}

		strcpy (dst, "; ");
		strcat (dst, op->comm);
	}
}

static void m68k_log_opcode(void *ext, unsigned long ir)
{
	struct m68k_instance *em = ext;
	uint32_t pc = e68_get_pc(em->e68);
	e68_dasm_t da;
	char str[256];

	if (em->log_flags & log_disasm) {		
		e68_dasm_mem(em->e68, &da, pc);
		m68k_dasm_str(str, &da);
		printf("m68k: [%08x] %s\n", pc, str);
	} else if (em->log_flags & log_prim_op)
		printf("m68k: [%08x] %04lx\n", pc, ir);
}

static void m68k_log_undef(void *ext, unsigned long ir)
{
	struct m68k_instance *em = ext;
	uint32_t pc = e68_get_pc(em->e68);

	printf("m68k: [%08x] UNDF %04lx\n", pc, ir);
}

static void m68k_log_exception(void *ext, unsigned tn)
{
	struct m68k_instance *em = ext;
	uint32_t pc = e68_get_pc(em->e68);

	printf("m68k: [%08x] EXCP %x\n", pc, tn);
}

static int m68k_tick(char *cd)
{
	vpiHandle			sys	= vpi_handle(vpiSysTfCall, 0);
	vpiHandle			argv	= vpi_iterate(vpiArgument, sys);
	vpiHandle			h, h_addr, h_data, h_flags;
	s_vpi_value			v;
	struct m68k_instance		*em;
	int				out_state;

	h = vpi_scan(argv);
	if (!h) {
		vpi_printf("m68k: Missing instance argument\n");
		goto fail;
	}
	v.format = vpiIntVal;
	vpi_get_value(h, &v);
	em = get_instance(v.value.integer);
	if (!em)
		goto fail;

	h_addr = vpi_scan(argv);
	h_data = vpi_scan(argv);
	h_flags = vpi_scan(argv);
	if (!h_addr || !h_data || !h_flags)
		goto fail;

	pthread_mutex_lock(&em->mutex);
	switch (em->state) {
	case emu_reset:
		em->state = emu_running;
	case emu_running:
		em->tick = 1;
		pthread_cond_signal(&em->cond_run);
	case emu_reading_done:
	case emu_writing_done:
		pthread_cond_wait(&em->cond_stop, &em->mutex);
	default:
		break;
	}

	switch (out_state = em->state) {
	case emu_writing:
		v.format = vpiIntVal;
		v.value.integer = em->data;
		vpi_put_value(h_data, &v, 0, vpiNoDelay);
	case emu_reading:
		v.format = vpiIntVal;
		v.value.integer = em->addr;
		vpi_put_value(h_addr, &v, 0, vpiNoDelay);
		v.format = vpiIntVal;
		v.value.integer = em->flags;
		vpi_put_value(h_flags, &v, 0, vpiNoDelay);
	default:
		break;
	}
	pthread_mutex_unlock(&em->mutex);

	v.format = vpiIntVal;
	v.value.integer = out_state;
	vpi_put_value(sys, &v, 0, vpiNoDelay);
 fail:
	return 0;
}

static int m68k_set_ipl(char *cd)
{
	vpiHandle			sys	= vpi_handle(vpiSysTfCall, 0);
	vpiHandle			argv	= vpi_iterate(vpiArgument, sys);
	vpiHandle			h, h_ipl;
	s_vpi_value			v;
	struct m68k_instance		*em;

	h = vpi_scan(argv);
	if (!h) {
		vpi_printf("m68k: Missing instance argument\n");
		goto fail;
	}
	v.format = vpiIntVal;
	vpi_get_value(h, &v);
	em = get_instance(v.value.integer);
	if (!em)
		goto fail;

	h_ipl = vpi_scan(argv);
	if (!h_ipl)
		goto fail;

	pthread_mutex_lock(&em->mutex);
	v.format = vpiIntVal;
	vpi_get_value(h_ipl, &v);
	e68_interrupt(em->e68, v.value.integer, 0, 1);
	pthread_mutex_unlock(&em->mutex);
 fail:
	return 0;
}

static int m68k_read_done(char *cd)
{
	vpiHandle			sys	= vpi_handle(vpiSysTfCall, 0);
	vpiHandle			argv	= vpi_iterate(vpiArgument, sys);
	vpiHandle			h, h_data;
	s_vpi_value			v;
	struct m68k_instance		*em;

	h = vpi_scan(argv);
	if (!h) {
		vpi_printf("m68k: Missing instance argument\n");
		goto fail;
	}
	v.format = vpiIntVal;
	vpi_get_value(h, &v);
	em = get_instance(v.value.integer);
	if (!em)
		goto fail;

	h_data = vpi_scan(argv);
	if (!h_data)
		goto fail;

	pthread_mutex_lock(&em->mutex);
	if (em->state != emu_reading)
		vpi_printf("m68k: m68k_read_done() in wrong state %d\n", em->state);
	else {
		v.format = vpiIntVal;
		vpi_get_value(h_data, &v);
		em->data = v.value.integer;
		em->state = emu_reading_done;
	}
	pthread_mutex_unlock(&em->mutex);

	pthread_cond_signal(&em->cond_run);
 fail:
	return 0;
}

static int m68k_write_done(char *cd)
{
	vpiHandle			sys	= vpi_handle(vpiSysTfCall, 0);
	vpiHandle			argv	= vpi_iterate(vpiArgument, sys);
	vpiHandle			h;
	s_vpi_value			v;
	struct m68k_instance		*em;

	h = vpi_scan(argv);
	if (!h) {
		vpi_printf("m68k: Missing instance argument\n");
		goto fail;
	}
	v.format = vpiIntVal;
	vpi_get_value(h, &v);
	em = get_instance(v.value.integer);
	if (!em)
		goto fail;

	pthread_mutex_lock(&em->mutex);
	if (em->state != emu_writing)
		vpi_printf("m68k: m68k_write_done() in wrong state %d\n", em->state);
	else
		em->state = emu_writing_done;
	pthread_mutex_unlock(&em->mutex);

	pthread_cond_signal(&em->cond_run);
 fail:
	return 0;
}

static void *em_start(void *arg)
{
	struct m68k_instance 	*em = arg;
	int need_reset = 1;

	/* The mutex is only unlocked when we are allowed to run */
	pthread_mutex_lock(&em->mutex);

	for (;;) {
		pthread_cond_signal(&em->cond_stop);
		pthread_cond_wait(&em->cond_run, &em->mutex);
		while(em->state == emu_running && em->tick) {
			if (need_reset)
				e68_reset(em->e68);
			else
				e68_clock(em->e68, 1);
			need_reset = 0;
			em->tick = 0;
		}
	}
	return NULL;
}

static int m68k_create(char *cd)
{
	struct m68k_instance 	*em;
	vpiHandle		sys	= vpi_handle(vpiSysTfCall, 0);
	vpiHandle		argv	= vpi_iterate(vpiArgument, sys);
	vpiHandle		h;
	s_vpi_value		ret;
	s_vpi_value		v;
	int			err;
	int			turbo = 1;
	int			log_flags = 0;

	ret.format = vpiIntVal;
	ret.value.integer = -1;

	if (m68k) {
		vpi_printf("m68k: Trying to create duplicate instance\n");
		goto fail;
	}

	h = vpi_scan(argv);
	if (h) {
		v.format = vpiIntVal;
		vpi_get_value(h, &v);
		turbo = !!v.value.integer;
	}
	vpi_printf("Turbo=%d\n", turbo);
	h = vpi_scan(argv);
	if (h) {
		v.format = vpiIntVal;
		vpi_get_value(h, &v);
		log_flags = v.value.integer;
	}

	m68k = em = calloc(sizeof(struct m68k_instance), 1);
	assert(em);
	em->e68 = e68_new();
	assert(em->e68);

	e68_set_mem_fct(em->e68, em,
			get_uint8, get_uint16, get_uint32,
			set_uint8, set_uint16, set_uint32);
	em->e68->turbo = turbo;
	em->log_flags = log_flags;
    	em->e68->log_ext = em;
	if (log_flags & (log_prim_op | log_disasm))
		em->e68->log_opcode = m68k_log_opcode;
	if (log_flags & log_undef)
		em->e68->log_undef = m68k_log_undef;
	if (log_flags & log_except)
		em->e68->log_exception = m68k_log_exception;
	em->e68->log_mem = NULL;

	em->state = emu_reset;

	pthread_mutex_init(&em->mutex, NULL);
	pthread_cond_init(&em->cond_run, NULL);
	pthread_cond_init(&em->cond_stop, NULL);
	pthread_mutex_lock(&em->mutex);
	err = pthread_create(&em->thread, NULL, em_start, em);
	if (err != 0) {
		ret.value.integer = -err;
		goto fail;
	}
	pthread_cond_wait(&em->cond_stop, &em->mutex);
	pthread_mutex_unlock(&em->mutex);
	ret.value.integer = 0x1234;

 fail:
	vpi_put_value(sys, &ret, 0, vpiNoDelay);

	return 0;
}

static void m68k_vpi_register()
{
	s_vpi_systf_data tf_data;

	tf_data.type      = vpiSysFunc;
	tf_data.tfname    = "$m68k_create";
	tf_data.calltf    = m68k_create;
	tf_data.compiletf = 0;
	tf_data.sizetf    = 0;
	tf_data.user_data = "$m68k_create";
	vpi_register_systf(&tf_data);

	tf_data.type      = vpiSysFunc;
	tf_data.tfname    = "$m68k_tick";
	tf_data.calltf    = m68k_tick;
	tf_data.compiletf = 0;
	tf_data.sizetf    = 0;
	tf_data.user_data = "$m68k_tick";
	vpi_register_systf(&tf_data);

	tf_data.type      = vpiSysTask;
	tf_data.tfname    = "$m68k_set_ipl";
	tf_data.calltf    = m68k_set_ipl;
	tf_data.compiletf = 0;
	tf_data.sizetf    = 0;
	tf_data.user_data = "$m68k_set_ipl";
	vpi_register_systf(&tf_data);

	tf_data.type      = vpiSysTask;
	tf_data.tfname    = "$m68k_read_done";
	tf_data.calltf    = m68k_read_done;
	tf_data.compiletf = 0;
	tf_data.sizetf    = 0;
	tf_data.user_data = "$m68k_read_done";
	vpi_register_systf(&tf_data);

	tf_data.type      = vpiSysTask;
	tf_data.tfname    = "$m68k_write_done";
	tf_data.calltf    = m68k_write_done;
	tf_data.compiletf = 0;
	tf_data.sizetf    = 0;
	tf_data.user_data = "$m68k_write_done";
	vpi_register_systf(&tf_data);
}

void (*vlog_startup_routines[])() = {
	m68k_vpi_register,
	0
};
